{-# LANGUAGE MultiParamTypeClasses, FlexibleInstances, TypeSynonymInstances,
             FlexibleContexts, ScopedTypeVariables #-}


{- Compilation to Term -}

module Code (Code(..)
            ,CVec(..)
            ,Comp
            ,compSys
            ,initState
            ) where
       

import StateCont
import Control.Monad
import Control.Monad.Identity
import Prelude hiding (div)
import Type
import Term
import Data
import Siso
import Vec

import Eval

{- Generate Term assembly code. -}
 
{- Code is parameterized with a phantom type to implement Data.  We
represent that type info as data in Term, making use of typeOf. This
keeps the Term data type simple, no need for GADTs etc.. -}

data Code t = Code { unCode :: Term }

instance Repr Code
instance MRepr Comp Code



{- State-Continuation monad used for Code interpretation.  State is
used for variable numbering, and tracking state, input, output
variables.  Continuation is used to implement Let-insertion. -}

type Comp = SC State Term

initState = (State 0 [])

regName n = "r" ++ show n


-- Primitive State operations
nextNameCounter   = SC $ \(State n s) k -> k (State (n + 1) s) n
saveStateVar v v0 = SC $ \(State n s) k -> k (State n ((v,v0):s)) ()

-- The continuation monad is used to insert Let expressions around the
-- current context.
primRef :: TypeOf t => Term -> Comp (Code t)
primRef prim = do
  (var, ref) <- genVarRef (undefined :: Code t)
  SC $ \s k ->    -- Modify continuation to
    Let var prim  -- 1. insert let binding around current context
    (k s ref)     -- 2. feed reference into current expression hole


-- Insert Fun
fun i = SC $ \s k -> Fun (flatten i) (k s i)

    
-- Fold loop

  
mFold :: (TypeOf e, TypeOf a) =>
         (Code e -> Code a -> Comp (Code a))
         -> Code a
         -> Code (CVec e)
         -> Comp (Code a)
mFold body (Code vec) (Code init) = do

  -- Gen refs
  (var_e, ref_e) <- genVarRef (undefined :: Code e)
  (var_a, ref_a) <- genVarRef (undefined :: Code a)
  SC $ \s k ->
    let (s', r) = compSC s $ body ref_e ref_a in
    Fold [Ref $ var_e, Ref $ var_a] r
    -- FIXME: ref_a is not correct.  Fish out the return value.
    (k s' ref_a)
  

-- State binding: initial values and updates.
state :: (Var -> Term -> Term -> Term) -> Var -> Code t -> Comp ()
state cons var (Code term) = do
  SC $ \s k ->     -- Modify continuation to
    cons var term  -- insert a state update assignment
    (k s ())

genVarRef :: TypeOf t => (Code t) -> Comp (Var, Code t)
genVarRef typEl = do
  num <- nextNameCounter
  let var = Var (typeOf typEl) (regName num) 
  return (var, Code $ Ref $ var)



-- Abstract literal value constructing and matching to be able to
-- write generic op1, op2 functions.
class Lit t' where
  lit   :: t'      -> Code t'
  unlit :: Code t' -> Maybe t'
instance Lit Treal where lit = Code . Lreal; unlit (Code (Lreal v)) = Just v; unlit _ = Nothing
instance Lit Tint  where lit = Code . Lint;  unlit (Code (Lint  v)) = Just v; unlit _ = Nothing
instance Lit Tbool where lit = Code . Lbool; unlit (Code (Lbool v)) = Just v; unlit _ = Nothing

-- Convert from Eval to Comp monad.
eval' :: Lit t => Eval (Val t) -> Comp (Code t)
eval' = return . lit . unVal . eval

-- Primitive operations all use Let insertion to replace expression
-- with variable reference.
op typ opc as = primRef $ Prim typ opc as


-- Evaluate literals, compile other.
op1 opc f a = rv where
  rv = ev $ unlit a
  ev (Just a') = eval' $ f (Val a')
  ev _         = op (typeOf $ unMRepr rv) opc [unCode a]

op2 opc f a b = rv where
  rv = ev (unlit a, unlit b)
  ev (Just a', Just b') = eval' $ f (Val a') (Val b')
  ev _                  = op (typeOf $ unMRepr rv) opc $ map unCode [a,b]
         


-- Data instances
instance Data Comp Code where  
  f2i = op1 "f2i" f2i
  i2f = op1 "i2f" i2f
  liti = lit
  litr = lit
  litb = lit

instance Ring Comp Code Treal where  
  add = op2 "add" add
  sub = op2 "sub" sub
  mul = op2 "mul" mul
  eq  = op2 "eq" eq
  lt  = op2 "lt" lt
  zero = lit 0
  one  = lit 1
  
instance Field Comp Code Treal where  
  div = op2 "div" div
  
instance Ring Comp Code Tint where  
  add = op2 "add" add
  sub = op2 "sub" sub
  mul = op2 "mul" mul
  eq  = op2 "eq" eq
  lt  = op2 "lt" lt
  zero = lit 0
  one  = lit 1



    
-- Keep track of a dictionary of hidden state variables for Sys compilation.
instance TypeOf t => SysState Comp (Code t) where
  genStateIn term@(Code init) = do
    (var, ref) <- genVarRef term
    state Prev var term
    -- saveStateVar var init
    return ref
  saveState (Code (Ref var), val) =
    state Next var val

-- Generate tupled input variables for Sys update function.
class MakeVar v where
  makeVar :: Comp v
instance TypeOf v => MakeVar (Code v) where
  makeVar = do
    (_, ref) <- genVarRef (undefined :: Code v)
    return ref
instance (MakeVar v1, MakeVar v2) => MakeVar (v1,v2) where
  makeVar = liftM2 (,) makeVar makeVar

class Flatten v where
  flatten :: v -> [Term]
instance TypeOf v => Flatten (Code v) where
  flatten (Code t) = [t]
instance (Flatten v1, Flatten v2) => Flatten (v1,v2) where
  flatten (v1,v2) = flatten v1 ++ flatten v2
                     

{- Generate input values for function, Sys. -}

-- Values that can be reduced to Comp values by feeding in variable
-- references.
  
class CompRed t o where
      compRed :: t -> Comp o

instance (MakeVar i, Flatten i) => 
         CompRed (i -> Comp o) -> o where
  compRed f = makeVar >>= fun >>= f
  
instance CompRed (Sys Comp i o) o where  
  compRef (Sys stateInit update) = do
    f <- compRed $ flip update
    stateVar <- genStateIn stateInit
    (stateNext, out) <- f stateVar
    saveState (stateVar, stateNext)
    return out




{- Main Compiler :: Sys -> Term

compResult is needed to fish out Context which has the compilation end
State, accessible when the final (return) continuation is called.

-}

{- Compile a monadic output, function, system. -}
compSC :: Flatten o =>
          State ->
          Comp o -> (State, Term)
compSC s_init sc = compResult $ cpsComp sc s_init k_ret where
  k_ret s_end rv = Context s_end $ Return $ flatten rv

compFun :: (MakeVar i, Flatten i, Flatten o) =>
           State ->
           (i -> Comp o) -> (State, Term)
compFun s f = compSC s $ appFun f

compSys :: (MakeVar i, Flatten i, Flatten o) =>
           State ->
           Sys Comp i o -> (State, Term)
compSys s sys = compSC s (appSys sys)


compResult :: Term -> (State, Term)
compResult = f where
  f (Context s t)  = (s, t)
  f (Fold ts  t1 t2) = (s, Fold ts t1 t2') where (s, t2') = f t2
  f (Let  v t1 t2)   = (s, Let  v  t1 t2') where (s, t2') = f t2
  f (Prev v t1 t2)   = (s, Prev v  t1 t2') where (s, t2') = f t2
  f (Next v t1 t2)   = (s, Next v  t1 t2') where (s, t2') = f t2
  f (Fun  ts t)      = (s, Fun  ts t')     where (s, t')  = f t                                                 

instance TypeOf t => TypeOf (Code t) where
  typeOf _ = typeOf (undefined :: t)

data CVec t = CVec t
instance Vec CVec
instance TypeOf t => TypeOf (CVec t) where
  typeOf _ = typeVec $ typeOf (undefined :: t)

  


instance (TypeOf e
         ,TypeOf a) => VecFold CVec Comp Code e a where
  vfold = mFold
