{-# LANGUAGE MultiParamTypeClasses, FlexibleContexts,
ExistentialQuantification, FlexibleInstances, TupleSections #-}
module Siso(Sys(..)
           ,sys
           ,Sig(..)
           ,SysState(..)
           ,mapSys
           ) where

import Prelude hiding ((.), id)
import Control.Category
import Control.Arrow
import Control.Monad.Identity


-- Below is a simplification of the ideas in Sys.hs

-- The missing link is to realize that 'm' is the compiler, and the
-- 's' type only needs to be specified up to its interaction with 'm'.

-- A state space model maps state and input to state and output.
-- Note: Using the form (s, i) -> m (s, o) creates too much friction.
type Siso m s i o = s -> i -> m (s, o)

-- Represent a system as initial value + update function.
-- State is kept abstract to allow this to fit the Category / Arrow instances.
data Sys m i o = forall s. (SysState m s) => Sys s (Siso m s i o)
type Sig m o = Sys m () o

-- Ignoring the SysState interface, this models a monadic stream map:
mapSys :: Monad m => Sys m i o -> [i] -> m [o] 
mapSys _ [] = return []
mapSys (Sys s f) (i:is) = do
  (s', o) <- f s i
  os <- mapSys (Sys s' f) is
  return $ o : os


-- The interaction between s and m can be kept mininal: enough to
-- generate a test value to feed into the update equation, and a
-- function to tuck away all state data/code into the compilation
-- monad.
class Monad m => SysState m s where
  -- Generate state input expression (e.g. variable) from (type of)
  -- state init.  Monad records initial value.
  genStateIn :: s -> m s
  
  -- Save:
  --  - state input expression as returned by genStateIn
  --  - next state expression
  saveState  :: (s, s) -> m () 




-- SysState instances for () and (,) used for state composition in
-- the Arrow instance.
instance (SysState m s1, SysState m s2) => SysState m (s1, s2) where

  genStateIn (init1, init2) =
    liftM2 (,) (genStateIn init1) (genStateIn init2)

  saveState ((var1, var2), (next1, next2)) = do
    saveState (var1, next1)
    saveState (var2, next2)

-- Instance for (), which arrises through arr lifting.
instance Monad m => SysState m () where
  genStateIn ()  = return () -- stub
  saveState _    = return () -- ignore stubs


-- Expose Sys composition as an Arrow

-- sys lifts a monadic function to a Sys.
sys :: Monad m => (i -> m o) -> Sys m i o
sys f = Sys () $ \() i -> do o <- f i; return ((), o)

-- arr only works for pure functions, which for Data are only
-- meta-operations (macros).
arr' f = sys $ return . f

  
instance Monad m => Category (Sys m) where

  id = arr' $ id

  -- Naming: msg: [m]onadic initial [s]tate for system [g]
  (Sys sg g) . (Sys sf f) = Sys (sf, sg) gf where
    gf (sf, sg) i = do
      (sf', x) <- f sf i
      (sg', o) <- g sg x
      return ((sf', sg'), o)
      
instance Monad m => Arrow (Sys m) where

  arr = arr'

  first (Sys sf f) = Sys sf f' where
    f' sf (i, x) = do
      (sf', o) <- f sf i
      return (sf', (o, x))



