{-# LANGUAGE MultiParamTypeClasses, FlexibleInstances, FunctionalDependencies #-}

{- Simple ADT for Comp.hs -}

module Term (Term(..)
            ,Var(..)
            ,State(..)
            ) where

import Control.Monad.Identity
import Type
import Data

type VarName = String
type OpName  = String

-- Ad-hoc expression language to represent a compiled form of Data.hs
-- This does not have a lot of structure constraints.
--
-- Code.hs produces a flat SSA-like language embedded in Term with:
--   Let has Prim as first Term argument, inner scope as second
--   Next has Ref or Lit as first Term argument, remainder of code as second
--   Prim has only Ref or Lit Terms
--   Ctx has only Ref or Lit Terms

data Term   = Ref     Var                              -- variable reference
            | Lbool   Tbool | Lint Tint | Lreal Treal  -- literal values
            | Prim    Type OpName [Term]               -- primitive operations
            | Let     Var Term Term                    -- variable binding
            | Prev    Var Term Term                    -- introduce state variable
            | Next    Var Term Term                    -- state update
            | Fold    [Term] Term Term                 -- stuff, body
            | Context State Term                       -- compilation state dump
            | Fun     [Term] Term
            | Return  [Term]                           -- computation result
            deriving (Eq)



-- Comp uses a state continuation monad with return value Term.  To
-- recover state, it needs to be embedded in the Term language, so
-- State is defined here.
data State = State {
  varCount    :: Int,
  stateVars   :: [(Var,Term)]  -- Term is initial value
  } deriving Eq


showAssign op v e c =
  show v ++ op ++ show e ++ "\n"
  
showLines :: Show x => String -> [x] -> String
showLines indent = concat . (map (\x -> indent ++ show x ++ "\n"))

instance Show Term where
  show (Fun  i b)     = "fun " ++ show i ++ "\n" ++ show b
  show (Let  v p c)   = show v ++ " = "  ++ show p ++ "\n" ++ show c
  show (Prev v i c)   = show v ++ " := " ++ show i ++ "\n" ++ show c
  show (Next v e c)   = show v ++ " <-"  ++ show e ++ "\n" ++ show c
  show (Ref v)        = show v
  show (Lbool l)      = show l
  show (Lint l)       = show l
  show (Lreal l)      = show l
  show (Prim t o ts)  = o ++ show ts
  show (Fold ts b n)  = "fold " ++ show ts ++ "\n" ++ show b ++ show n
  show (Return rv)    = "ret "  ++ show rv ++ "\n"
    

  show (Context s t) = 
    "context:" ++ show s ++ show t

instance Show State where
  show (State c s) =
    "\n  count: " ++ show c ++
    "\n  state: " ++ show (reverse s) ++ "\n"
  

-- Variables can refer to data values or functions.  In the latter
-- case type refers to the return value's type.
data Var    = Var {varType :: Type, varName :: VarName }
            deriving (Eq)
                     
instance Show Var where                     
  show (Var (Type typename dim) name) = showDim dim ++ name

showDim 0 = ""
showDim n = "^" ++ (showDim $ n - 1)
                               

    
